import re
import ast
from textwrap import dedent
from functools import lru_cache

function_wrong_symbols = re.compile(r"[\ \"\'\:;\<>{}[\]~`]")
params_wrong_symbols = re.compile(r'\w*\s*\(\s*\w*\s*\)\s*')
string_match = re.compile(r'"|\'\w*\"|\'')


@lru_cache(maxsize=4096)
def parse_func_str(val, auto_quote=True):
    """
    Parse function call from string

    Returns:
        tuple fn, args, kwargs
    Raises:
        ValueError: if string is invalid
    """

    def _invalid_arg(arg):
        raise ValueError('Invalid argument')

    def _format_arg(arg, kw=False):
        try:
            return ast.literal_eval(arg.value if kw else arg)
        except:
            if auto_quote:
                try:
                    return [
                        _format_arg(a)
                        for a in (arg.value.elts if kw else arg.elts)
                    ]
                except:
                    try:
                        return '{}.{}'.format(
                            arg.value.value.id,
                            arg.value.attr) if kw else '{}.{}'.format(
                                arg.value.id, arg.attr)
                    except:
                        try:
                            return arg.value.id if kw else arg.id
                        except:
                            _invalid_arg(arg)
            else:
                _invalid_arg(arg)

    val = val.strip()
    if not val or not val.endswith(')'):
        raise ValueError('Invalid syntax')
    params = val[val.find('(') + 1:val.rfind(')')].strip()
    fn = val[:val.find('(')].strip()
    if fn[0] in ('@', '!', '?'):
        pfx = fn[0]
        fn = fn[1:]
    else:
        pfx = ''
    # check function name
    if not fn or function_wrong_symbols.search(fn):
        raise ValueError('Invalid symbols in function name')
    # try parsing params with ast
    try:
        code = 'f({})'.format(params)
        try:
            tree = ast.parse(code)
        except:
            raise ValueError('Invalid syntax')
        funccall = tree.body[0].value
        args = [_format_arg(arg) for arg in funccall.args]
        kwargs = {
            arg.arg: _format_arg(arg, kw=True) for arg in funccall.keywords
        }
    except ValueError:
        # try parsing manually, less features
        try:
            if val.rsplit(')', 1)[1].strip():
                raise ValueError
        except IndexError:
            raise ValueError
        params = [v for v in val.split('(', 1)[1].rsplit(')', 1)[0].split(',')]
        new_params = []
        if auto_quote:
            for p in params:
                new_p = p.split('=') if p.__contains__('=') else [None, p]
                new_p1 = new_p[1].strip()
                if params_wrong_symbols.match(new_p[1].strip()):
                    raise ValueError(
                        'Invalid symbols in args - {}'.format(new_p1))
                if new_p1 and not string_match.match(new_p1):
                    try:
                        float(new_p1)
                    except ValueError:
                        p = '"{}"'.format(new_p[1].strip(
                        )) if new_p[0] is None else '{}="{}"'.format(
                            new_p[0], new_p1)
                new_params.append(p)
        else:
            new_params = params
        code = dedent("""
                def {}(*args, **kwargs):
                    global a, kw
                    a = args
                    kw = kwargs
                {}""").format(fn, '{}({})'.format(fn, ', '.join(new_params)))
        d = {}
        try:
            exec(code, d)
            args = list(d['a'])
            kwargs = d['kw']
        except Exception as e:
            raise ValueError(str(e))
    return pfx + fn, args, kwargs



import threading
import logging

from functools import wraps

class Locker:
    """
    Function locker decorator object
    """

    def __init__(self, mod='', timeout=5, relative=True):
        """
        Args:
            mod: module to report in logs
            timeout: lock timeout (default: 5 sec)
            relative: use thread-relative locking (default: True)
        """
        self.lock = threading.RLock() if relative else threading.Lock()
        self.mod = '' if not mod else mod + '/'
        self.relative = relative
        self.timeout = timeout

    def __call__(self, f):

        @wraps(f)
        def do(*args, **kwargs):
            if not self.lock.acquire(timeout=self.timeout):
                logging.critical('{}{} locking broken'.format(
                    self.mod, f.__name__))
                self.critical()
                return None
            try:
                return f(*args, **kwargs)
            finally:
                self.lock.release()

        return do

    def critical(self):
        """
        Function is called when lock can not be obtained

        Empty by default, override or monkey-patch
        """
        pass



import importlib.util
from functools import lru_cache
from types import SimpleNamespace
from pathlib import Path


class ResourceStorage:
    """
    Resource storage loader

    Usage example:

    from pyaltt.res import ResourceStorage
    from functools import partial

    rs = ResourceStorage(mod=mymod)

    rq = partial(rs.get, resource_subdir='sql', ext='sql')

    rq('object.select.data') - will load resource from (will try all
    variations until file is found):

    * sql/object.select.data.sql
    * sql/object/select.data.sql
    * sql/object/select/data.sql

    Resource is loaded once and cached forever.
    """

    def __init__(self, resource_dir='.', mod=None):
        """
        init resource storage for module

        If module is specified, set directory to module_dir/resources

        Args:
            resource_dir: resource directory or
            mod: module name
        Raises:
            LookupError: if module name is specified, but module is not found
        """
        if mod:
            spec = importlib.util.find_spec(mod)
            if spec is None:
                raise LookupError
            else:
                self.resource_dir = (Path(spec.origin).absolute().parent /
                                     'resources').as_posix()
        else:
            self.resource_dir = resource_dir

    @lru_cache(maxsize=None)
    def get(self,
            resource_id,
            resource_subdir=None,
            ext=None,
            mode='r',
            default=None):
        """
        Get resource

        Loads resource from resource storage directory

        Args:
            resource_id: resource id. dots are replaced with "/" automatically
                until resource is found
            resource_subdir: resource sub directory
            ext: resource extension
            mode: file open mode (default: "r", use "rb" for binary data)
            default: return default value if resource is not found
        Raises:
            LookupError: if resource is not found and no default value provided
        """
        fname = str(resource_id)
        while True:
            fname_full = '{}{}/{}{}'.format(
                self.resource_dir,
                '' if resource_subdir is None else f'/{resource_subdir}', fname,
                '' if ext is None else f'.{ext}')
            try:
                with open(fname_full, mode) as fh:
                    return fh.read()
            except FileNotFoundError:
                if '.' not in fname: break
                fname = fname.replace('.', '/', 1)
        if default is None:
            raise LookupError('resource not found: {}{}'.format(
                resource_id, '' if ext is None else f'.{ext}'))
        else:
            return default



# dirty fix for cryptography.Random
import time
try:
    getattr(time, 'clock')
except AttributeError:
    time.clock = time.time

def gen_random_str(length=32):
    """
    Generate random string (letters+numbers)

    Args:
        length: string length (default: 32)
    """
    import string
    import random
    symbols = string.ascii_letters + '0123456789'
    return ''.join(random.choice(symbols) for i in range(length))


def encrypt(raw, key, hmac_key=None, key_is_hash=False, b64=True, bits=256):
    """
    Encrypt bytes with AES-CBC

    Args:
        raw: bytes to encrypt
        key: encryption key
        hmac_key: HMAC key (optional), True or custom key
        key_is_hash: consider encryption key is sha256 hash
        b64: encode result in base64 (default: True)
        bits: key size (128, 192 or 256, default is 256)
    Returns:
        encrypted block + 32-byte HMAC signature (if hmac_key is specified)

    note: if hmac_key is True and key is hash, sha512 sum is required
    """
    from Crypto.Cipher import AES
    from Crypto import Random
    import hashlib
    import hmac
    if isinstance(raw, str):
        raw = raw.encode()
    if hmac_key is True:
        h = key if key_is_hash else hashlib.sha512(
            key.encode() if isinstance(key, str) else key).digest()
        keyhash = h[:bits // 8]
        hmac_key = h[:-32]
    else:
        keyhash = key if key_is_hash else hashlib.sha256(
            key.encode() if isinstance(key, str) else key).digest()[:bits // 8]
    length = 16 - (len(raw) % 16)
    raw += bytes([length]) * length
    iv = Random.new().read(AES.block_size)
    cipher = AES.new(keyhash, AES.MODE_CBC, iv)
    val = iv + cipher.encrypt(raw)
    if hmac_key:
        val += hmac.new(
            hmac_key.encode() if isinstance(hmac_key, str) else hmac_key, val,
            hashlib.sha256).digest()
    if b64:
        import base64
        return base64.b64encode(val).decode()
    else:
        return val


def decrypt(enc, key, hmac_key=None, key_is_hash=False, b64=True, bits=256):
    """
    Decrypt encoded data with AES-CBC

    Args:
        enc: data to decrypt
        key: decryption key
        key_is_hash: consider decryption key is sha256 hash
        hmac_key: HMAC key (optional), True or custom key
        b64: decode data from base64 (default: True)
        bits: key size (128, 192 or 256, default is 256)
    Raises:
        ValueError: if HMAC auth failed

    note: if hmac_key is True and key is hash, sha512 sum is required
    """
    from Crypto.Cipher import AES
    import hashlib
    import hmac
    if b64:
        import base64
        enc = base64.b64decode(enc)
    if hmac_key is True:
        h = key if key_is_hash else hashlib.sha512(
            key.encode() if isinstance(key, str) else key).digest()
        keyhash = h[:bits // 8]
        hmac_key = h[:-32]
    else:
        keyhash = key if key_is_hash else hashlib.sha256(
            key.encode() if isinstance(key, str) else key).digest()[:bits // 8]
    iv = enc[:16]
    if hmac_key:
        if hmac.new(
                hmac_key.encode() if isinstance(hmac_key, str) else hmac_key,
                enc[:-32], hashlib.sha256).digest() != enc[-32:]:
            raise ValueError('HMAC auth failed')
    cipher = AES.new(keyhash, AES.MODE_CBC, iv)
    data = cipher.decrypt(enc[16:-32 if hmac_key else None])
    return data[:-data[-1]]


class Rioja:
    """
    Rioja (ˈrjoxa) is a crypto engine, similar to Fernet, but:

    - implements AES-CBC-HMAC up to AES256 (default)
    - more simple to use
    """

    def __init__(self, key, bits=256):
        """
        Args:
            key: encryption key
            bits: key size (128, 192 or 256, default is 256)
        """
        import hashlib
        self.__keyhash = hashlib.sha512(
            key.encode() if isinstance(key, str) else key).digest()
        self.__bits = bits

    def encrypt(self, raw, b64=True):
        """
        Args:
            raw: bytes to encrypt
            b64: encode result in base64 (default: True)
        """
        return encrypt(raw,
                       self.__keyhash,
                       hmac_key=True,
                       key_is_hash=True,
                       b64=b64,
                       bits=self.__bits)

    def decrypt(self, enc, b64=True):
        """
        Args:
            enc: data to decrypt
            b64: decode data from base64 (default: True)
        Raises:
            ValueError: if HMAC auth failed
        """
        return decrypt(enc,
                       self.__keyhash,
                       hmac_key=True,
                       key_is_hash=True,
                       b64=b64,
                       bits=self.__bits)

default_public_key = None
"""
default per-project public key for signature verification
"""

def sign(content, private_key, key_password=None):
    """
    Sign content with RSA key
    Args:
        content: content to sign
        private_key: private RSA (PEM) key
        key_password: key password (optional)
    Returns:
        base64-encoded RSA signature
    """
    from cryptography.hazmat.primitives import serialization, hashes
    from cryptography.hazmat.primitives.asymmetric import padding
    from cryptography.hazmat.backends import default_backend
    import hashlib
    import base64

    if not isinstance(content, bytes):
        content = str(content).encode()
    if not isinstance(private_key, bytes):
        private_key = str(private_key).encode()

    privkey = serialization.load_pem_private_key(private_key,
                                                 password=key_password,
                                                 backend=default_backend())

    prehashed = hashlib.sha256(content).digest()

    sig = privkey.sign(
        prehashed,
        padding.PSS(mgf=padding.MGF1(hashes.SHA256()),
                    salt_length=padding.PSS.MAX_LENGTH), hashes.SHA256())

    return base64.b64encode(sig).decode()


def verify_signature(content, signature, public_key=None):
    """
    Verify content with RSA signature

    If public key is not specified, global per-project key is used

    Args:
        content: content to sign
        signature: base64-encoded RSA signature
        public_key: public RSA (PEM) key
    Returns:
        base64-encoded RSA signature
    """
    import hashlib
    import base64
    from cryptography.exceptions import InvalidSignature
    from cryptography.hazmat.primitives import serialization, hashes
    from cryptography.hazmat.primitives.asymmetric import padding
    from cryptography.hazmat.backends import default_backend

    if not isinstance(content, bytes):
        content = str(content).encode()
    if public_key is None:
        public_key = default_public_key
    elif not isinstance(public_key, bytes):
        public_key = str(public_key).encode()

    pubkey = serialization.load_pem_public_key(public_key,
                                               backend=default_backend())
    prehashed_msg = hashlib.sha256(content).digest()
    decoded_sig = base64.b64decode(signature)
    pubkey.verify(
        decoded_sig, prehashed_msg,
        padding.PSS(mgf=padding.MGF1(hashes.SHA256()),
                    salt_length=padding.PSS.MAX_LENGTH), hashes.SHA256())
    return True



"""
Extra mods required: netaddr
"""

def parse_host_port(hp, default_port=0):
    """
    Parse host/port from string

    Doesn't work with IPv6

    Args:
        hp: host/port string to parse
        default_port: default port if no port is specified (default: 0)

    Returns:
        tuple (host, port)
    """
    if hp.find(':') == -1:
        return (hp, default_port)
    else:
        host, port = hp.rsplit(':', 1)
        return (host, int(port))


def netacl_match(ip, acl):
    """
    Check if IP mathches network ACL

    Doesn't work with IPv6

    Args:
        ip: IP address to check
        acl: list of netadd.IPNetwork objects

    Returns:
        True if ACL matches, False if not
    """
    from netaddr import IPAddress
    ipa = IPAddress(ip)
    for a in acl:
        if ipa in a: return True
    return False


def generate_netacl(source, default=[]):
    """
    Generates IP ACL

    Doesn't work with IPv6

    Args:
        source: string with IP/network or iterable
        default: returns if source is empty
    """
    if source:
        from netaddr import IPNetwork
        return [
            IPNetwork(h)
            for h in ([source] if isinstance(source, str) else source)
        ]
    else:
        return default



from pyaltt2.network import parse_host_port
import smtplib


class SMTP:
    """
    SMTP sender
    """

    def __init__(self,
                 host='localhost',
                 port=25,
                 tls=False,
                 ssl=False,
                 login=None,
                 password=None,
                 debug=False):
        """
        Args:
            host: SMTP host
            port: SMTP port
            tls: use TLS (default: False)
            ssl: use SSL (default: False)
            login: SMTP login
            password: SMTP password
        """
        self.host = host
        self.port = port
        self.tls = tls
        self.ssl = ssl
        self.login = login
        self.password = password
        self.sendfunc = smtplib.SMTP_SSL if ssl else smtplib.SMTP
        self.debug = debug

    def send(self, sender, rcpt, body):
        """
        Send raw email

        Args:
            sender: E-Mail sender
            rcpt: E-Mail recipient
            body: E-Mail body (string)
        """
        sm = self.sendfunc(self.host, self.port)
        if self.debug:
            sm.set_debuglevel(9)
        sm.ehlo()
        if self.tls:
            sm.starttls()
        if self.login is not None:
            sm.login(self.login, self.password)
        sm.sendmail(sender, rcpt, body)
        sm.close()

    def sendmail(self, sender, rcpt, subject='', text='', html=None):
        """
        Send text/html email

        Args:
            sender: E-Mail sender
            rcpt: E-Mail recipient
            subject: E-Mail subject
            text: message text in plain
            html: message text in HTML
        """
        from email.mime.multipart import MIMEMultipart
        from email.mime.text import MIMEText
        if html:
            m = MIMEMultipart('alternative')
            part_text = MIMEText(text, 'plain')
            part_html = MIMEText(html, 'html')
            m.attach(part_text)
            m.attach(part_html)
        else:
            m = MIMEText(text)
        m['Subject'] = subject
        m['From'] = sender
        m['To'] = rcpt[0] if isinstance(rcpt, list) and len(rcpt) > 1 else rcpt
        self.send(sender, rcpt, m.as_string())



"""
Extra mods required: pyyaml, jsonschema
"""
import yaml
import os

try:
    yaml.warnings({'YAMLLoadWarning': False})
except:
    pass


def load_yaml(fname, schema=None):
    """
    Load config from YAML/JSON file

    Args:
        fname: file name to load
        schema: JSON schema for validation
    """
    with open(fname) as fh:
        data = yaml.load(fh.read())
    if schema:
        import jsonschema
        jsonschema.validate(data, schema=schema)
    return data


def config_value(env=None,
                 config=None,
                 config_path=None,
                 to_str=False,
                 read_file='r',
                 in_place=False,
                 default=LookupError):
    """
    Get config value

    Args:
        env: search in system env, if specified, has top priority
        config: config dict to process
        config_path: config path (e.g. /some/long/key)
        to_str: strinify value before returning
        read_file: if value starts with '/', read it from file with mode
            (default: 'r')
        in_place: replace value in config dict
        default: default value
    Raises:
        LookupError: if value is not found and no default value specified
    """
    value = default
    if config_path is not None and config_path.startswith('/'):
        config_path = config_path[1:]
    if env is not None and env in os.environ:
        value = os.environ[env]
    if value is default or in_place and \
            (config is not None and config_path is not None):
        path = config_path.split('/')
        x = config
        for p in path[:-1]:
            if p not in x and in_place:
                x[p] = {}
            x = x.get(p, {})
        if value is default:
            if path[-1] in x:
                value = x[path[-1]]
    if value is LookupError:
        if env is None and config is not None and config_path is not None:
            msg = f'unable to find value of {config_path}'
        elif env is not None and (config is None or config_path is None):
            msg = f'OS variable {env} is not set'
        elif env is not None and config is not None and config_path is not None:
            msg = (f'unable to find value of {config_path}, '
                   f'OS variable {env} is not set as well')
        else:
            msg = None
        raise LookupError(msg)
    elif read_file and isinstance(value, str) and (value.startswith('/') or
                                                   value.startswith('./')):
        try:
            with open(str(value), read_file) as fh:
                value = fh.read()
        except:
            raise LookupError(f'Unable to read value from {str(value)}')
    elif to_str:
        value = str(value)
    if in_place:
        x[path[-1]] = value
    return value


def choose_file(fname=None, env=None, choices=[]):
    """
    Chooise existing file

    Returned file path is user-expanded

    Args:
        fname: if specified, has top priority and others are not chechked
        env: if specified and set, has second-top priority and choices are not
            inspected
        choices: if env is not set or not specified, choose existing file from
            the list
    Raises:
        LookupError: if file doesn't exists
    """
    if fname:
        fname = os.path.expanduser(fname)
        if os.path.exists(fname):
            return fname
        else:
            raise LookupError(f'No such file {fname}')
    elif env and env in os.environ:
        fname = os.path.expanduser(os.environ[env])
        if os.path.exists(fname):
            return fname
        else:
            raise LookupError(f'No such file {env} = {fname}')
    else:
        ch = []
        for c in choices:
            fname = os.path.expanduser(str(c))
            if os.path.exists(fname):
                return fname
            else:
                ch.append(fname)
        raise LookupError('File not found (tried {}{})'.format(
            env + (', ' if choices else '') if env else '', ', '.join(ch)))
